using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace ZBufferDemo
{
	public class MainForm : System.Windows.Forms.Form
	{
		private const int numberRingSegments = 20;
		private const int numberTubeSegments = 10;
		private const int numberTriangles = 2*numberRingSegments*numberTubeSegments;
		
		private Device device;
		private VertexBuffer myVertices;
		private Random myRandom = new Random();
		private Matrix worldMatrix = Matrix.Identity;
		private int numberTrianglesToDraw = numberTriangles;
		private bool useZBuffer = true;

		private System.ComponentModel.Container components = null;

		public MainForm()
		{
			InitializeComponent();

			SetStyle(ControlStyles.Opaque, true);

			PresentParameters pp = new PresentParameters();
			pp.Windowed = true;
			pp.SwapEffect = SwapEffect.Discard;
			pp.AutoDepthStencilFormat = DepthFormat.D24X8;
			pp.EnableAutoDepthStencil = true;
			device = new Device(0, DeviceType.Hardware, this, CreateFlags.HardwareVertexProcessing, pp);

			myVertices = buildTorus(device, 1.0, numberRingSegments, 0.6, numberTubeSegments);
		}

		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Vom Windows Form-Designer generierter Code
		/// <summary>
		/// Erforderliche Methode fr die Designeruntersttzung. 
		/// Der Inhalt der Methode darf nicht mit dem Code-Editor gendert werden.
		/// </summary>
		private void InitializeComponent()
		{
			// 
			// MainForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(424, 374);
			this.Name = "MainForm";
			this.Text = "Z-Buffer Demo";

		}
		#endregion

		[STAThread]
		static void Main() 
		{
			Application.Run(new MainForm());
		}
	
		protected override void OnPaint(PaintEventArgs e)
		{
			device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.White, 1.0f, 0);
			device.BeginScene();
				
			device.Transform.World = worldMatrix;                            
			device.Transform.View = Matrix.LookAtLH(new Vector3(0.0f, 0.0f, -7.0f), new Vector3(), new Vector3(0.0f, 1.0f, 0.0f));
			device.Transform.Projection = Matrix.PerspectiveFovLH(Geometry.DegreeToRadian(20.0f), ClientSize.Width/(0.001f+ClientSize.Height), 0.1f, 50.0f);
				
			device.RenderState.ZBufferEnable = useZBuffer;
				
			device.SetStreamSource(0, myVertices, 0); 
			device.VertexFormat = CustomVertex.PositionColored.Format; 
			device.RenderState.Lighting = false; 
			device.RenderState.CullMode = Cull.None;
			device.DrawPrimitives(PrimitiveType.TriangleList, 0, numberTrianglesToDraw);
				
			device.EndScene();
			device.Present();
			Invalidate();
		}

		protected override void OnKeyDown(KeyEventArgs e)
		{
			e.Handled = true;
			switch(e.KeyCode)
			{
				case Keys.Escape:
					this.Close();
					break;
				case Keys.PageUp:
					if(numberTrianglesToDraw < numberTriangles)
					{
						numberTrianglesToDraw++;
					}
					break;
				case Keys.PageDown:
					if(numberTrianglesToDraw > 1)
					{
						numberTrianglesToDraw--;
					}
					break;
				case Keys.D:
					useZBuffer = !useZBuffer;
					break;
				case Keys.Up:
					worldMatrix = worldMatrix*Matrix.RotationX( 0.1f);
					break;
				case Keys.Down:
					worldMatrix = worldMatrix*Matrix.RotationX(-0.1f);
					break;					
				case Keys.Left:
					worldMatrix = worldMatrix*Matrix.RotationY( 0.1f);
					break;
				case Keys.Right:
					worldMatrix = worldMatrix*Matrix.RotationY(-0.1f);
					break;
				default:
					e.Handled = false;
					break;
			}
			base.OnKeyDown(e);
		}

		private VertexBuffer buildTorus(Device device, double ringRadius, int ringSegments, double tubeRadius, int tubeSegments)
		{
			// Better do this using an additional IndexBuffer, which we'll cover later.
			
			VertexBuffer vb = new VertexBuffer(typeof(CustomVertex.PositionColored), 6*ringSegments*tubeSegments, device, Usage.WriteOnly, 0, Pool.Managed);
			CustomVertex.PositionColored[] v = (CustomVertex.PositionColored[]) vb.Lock(0, 0);
			for(int ringIndex = 0; ringIndex < ringSegments; ringIndex++)
			{
				for(int tubeIndex = 0; tubeIndex < tubeSegments; tubeIndex++)
				{
					int baseIndex = 6*(tubeIndex+tubeSegments*ringIndex);
					
					// one quadrangular face
					Vector3 a = torusPoint(ringRadius, ringSegments, ringIndex,
						tubeRadius, tubeSegments, tubeIndex);
					Vector3 b = torusPoint(ringRadius, ringSegments, ringIndex + 1,
						tubeRadius, tubeSegments, tubeIndex);
					Vector3 c = torusPoint(ringRadius, ringSegments, ringIndex + 1,
						tubeRadius, tubeSegments, tubeIndex + 1);					
					Vector3 d = torusPoint(ringRadius, ringSegments, ringIndex,
						tubeRadius, tubeSegments, tubeIndex + 1);					
					
					// build quadrangle as two triangles
					v[baseIndex    ].Position = a;
					v[baseIndex + 1].Position = b;
					v[baseIndex + 2].Position = d;
					v[baseIndex + 3].Position = b;
					v[baseIndex + 4].Position = c;
					v[baseIndex + 5].Position = d;
					
					v[baseIndex    ].Color
						= v[baseIndex + 1].Color
						= v[baseIndex + 2].Color
						= myRandom.Next(0xFFFFFF);
					
					v[baseIndex + 3].Color
						= v[baseIndex + 4].Color
						= v[baseIndex + 5].Color
						= myRandom.Next(0xFFFFFF);					
				}
			}
			vb.Unlock();
			return vb;
		}
		
		private Vector3 torusPoint(double ringRadius, int ringSegments, int ringIndex,
			double tubeRadius, int tubeSegments, int tubeIndex)
		{
			double ringAngle = 2.0*Math.PI*ringIndex/ringSegments;
			double tubeAngle = 2.0*Math.PI*tubeIndex/tubeSegments;
			double radialDistance = ringRadius + tubeRadius*Math.Cos(tubeAngle);
			return new Vector3((float)(radialDistance * Math.Cos(ringAngle)),
				(float)(tubeRadius * Math.Sin(tubeAngle)),
				(float)(radialDistance * Math.Sin(ringAngle))
				);
		}
	}
}
